from atexit import register
import os 

_config_dir = ".UCQ_config"
_layout_file = os.path.join(_config_dir, "layout.json")
_trigger_file = ".trigger"
_master_show = False
_show_plt = False
_states = {}
_circs = []
_hists = []


# cleans up the config directory on init of this python module
if _config_dir in os.listdir():
    _master_show = True
    #print("cleaning up config dir")
    for item in os.listdir(_config_dir):
        # deletes png html or the trigger file from the config dir
        if item.endswith(".png") or item.endswith(".html") or item == _trigger_file:
            os.remove(os.path.join(_config_dir, item))

def _exit():
    for item in os.listdir(_config_dir):
        # deletes png html or the trigger file from the config dir
        if item == _trigger_file:
            os.remove(os.path.join(_config_dir, item))

register(_exit)


from qiskit import QuantumCircuit, Aer, execute
from qiskit.quantum_info import Statevector
from qiskit.visualization import plot_histogram
import matplotlib.pyplot as plt
from atexit import register
from math import log
from . import _states, _circs, _hists, _master_show, _show_plt
from ._src import get_path

_circ_count = 0
_state_count = 0
_hist_count = 0

def _show_at_exit():
    global _show_plt
    if _show_plt:
        #print("opening mpl figures")
        plt.show()

register(_show_at_exit)

# diplays the image in the viewer or saves the image to the inputted path
def display(circuit:QuantumCircuit, path:str=""):
    global _circ_count, _circs, _master_show, _show_plt
    circuit.draw(output='mpl')
    plt.tight_layout()
    if len(path): plt.savefig(path)
    elif _master_show:
        #print("displaying circuit")
        p = get_path(f"_circ_{_circ_count}.png")
        plt.savefig(p)
        _circs.append(p)
        _circ_count+=1
    else: 
        _show_plt = True

# generates binary strings
def getbin(n, s=['']):
    global _config
    if n > 0: return [*getbin(n - 1, [i + '0' for i in s]), *getbin(n - 1, [j + '1' for j in s])]
    return s

# displays the statevector of the circuit and can return it
def state(circuit:QuantumCircuit, show=True):
    global _state_count, _states, _master_show, _show_plt
    _state = Statevector.from_instruction(circuit).data
    _num_bits = int(log(len(_state))/log(2))
    
    _options = getbin(_num_bits)
    _data = {}
    for i in range(len(_state)):
        val = _state[i]
        if type(val) == complex:
            val = round(val.real, 10) + round(val.imag, 10) *1j
        else:
            val = round(val, 10)
        _data[_options[i]] = str(val).replace("(", "").replace(")", "")

    if show and _master_show:
        #print("showing state vector")
        if len(_states):
            if len(_options[i]) > len(list(_states.keys())[0]):
                raise KeyError("States must be obtained from the same circuit")
            for item in _data:
                _states[item].append(_data[item])
        else:
            for item in _data:
                _states[item] = [_data[item]]

    return _state

# displays the histogram of the circuit after execution in the viewer
def counts(circuit:QuantumCircuit, backend=Aer.get_backend('qasm_simulator'), path:str=""):
    global _hist_count, _hists, _master_show, _show_plt
    counts = execute(circuit, backend=backend, shots=1024).result().get_counts()
    if len(path): plt.savefig(path)
    elif _master_show:
        #print("displaying histogram")
        plot_histogram(counts)
        p = get_path(f"_hist_{_hist_count}.png")
        plt.savefig(p)
        _hists.append(p)
        _hist_count+=1
    else:
        plot_histogram(counts)
        _show_plt = True
    return counts





from . import _states, _circs, _hists, _layout_file, _master_show
from ._src import _trigger
from atexit import register
import os, json

_layout = {}
_to_flip = {}

# turns path into absolute path if it isn't
def abs_path(path): return os.path.abspath(path.replace("~", os.path.expanduser("~")))

# converts list of image files to html img elements
def image_list_to_str(image_list:list[str])->str:
    to_return = ""
    for item in image_list:
        to_return+=f"<img src=\"{{URI}}{abs_path(item)}\" alt=\"no image to display\">"
    return to_return

def _inverter(layout, flip:dict[str : str]):
    for item in list(layout):
        if item in flip:
            other = layout[item]
            layout[item] = layout[flip[item]]
            layout[flip[item]] = other

    
    for item in list(layout):
        if not isinstance(layout[item], str):
            layout[item] = _inverter(layout[item], flip)

    return layout

def invert():
    global _to_flip
    _to_flip = {"left" : "right", "top" : "bottom"}

def horizontal_invert():
    global _to_flip
    _to_flip = {"left" : "right"}


def vertical_invert():
    global _to_flip
    _to_flip = {"top" : "bottom"}


# default layout of the viewer
def default():
    global _layout, _states, _circs, _hists
    # if the statevector and an image is to be rendered
    if len(_states) and (len(_hists) or len(_circs)):
        #state_path = abs_path(os.path.join(_config_dir,  "_state_.html"))
        msg = "\\[\\begin{matrix} "
        length = len(_states)
        for i, item in enumerate(list(_states)):
            if i == 0:
                msg += ("\\text{bits}")
                for j in range(len(_states[item])):
                    msg += f" & \\text{{call {j+1}}}"
                msg += "\\\\"
            if i < length - 1:
                msg+=(f"{item} & " + "&".join(_states[item]) + "\\\\")
            else:
                msg+=(f"{item} & " + "&".join(_states[item]))
        msg+="\\end{matrix}\\]"
        _layout["left"] = msg #f"<div data-include=\"{{URI}}{state_path}\"></div>"

        if len(_hists) and len(_circs):
            _layout["right"] = {"top" : image_list_to_str(_circs), "bottom" : image_list_to_str(_hists)}
        elif len(_hists):
            _layout["right"] = image_list_to_str(_hists)
        elif len(_circs):
            _layout["right"] = image_list_to_str(_circs)
    
    elif len(_states):
        #state_path = abs_path(os.path.join(_config_dir,  "_state_.html"))
        msg = "\\[\\begin{matrix} "
        length = len(_states)
        for i, item in enumerate(list(_states)):
            if i == 0:
                msg += ("\\text{bits}")
                for j in range(len(_states[item])):
                    msg += f" & \\text{{call {j+1}}}"
                msg += "\\\\"
            if i < length - 1:
                msg+=(f"{item} & " + "&".join(_states[item]) + "\\\\")
            else:
                msg+=(f"{item} & " + "&".join(_states[item]))
        msg+="\\end{matrix}\\]"
        _layout["only"] = msg #f"<div data-include=\"{{URI}}{state_path}\"></div>"

    elif len(_hists) or len(_circs):
        if len(_hists) and len(_circs):
            _layout["top"] = image_list_to_str(_circs)
            _layout["bottom"] = image_list_to_str(_hists)
        elif len(_hists):
            _layout["only"] = image_list_to_str(_hists)
        elif len(_circs):
            _layout["only"] = image_list_to_str(_circs)
    
    else:
        _layout["only"] = "<h1>No data to display</h1>"

def _run():
    global _layout, _to_flip, _states, _circs, _hists, _master_show
    if _master_show:
        # running the default layout generator
        default()
        if len(_to_flip):
            _layout = _inverter(_layout, _to_flip)

        #print("unloading layout")
        with open(_layout_file, 'w') as f:
            f.write(json.dumps(_layout, indent=2))
        
        # clearing the values
        _to_flip = {}
        _layout = {}
        _states = []
        _circs = [] 
        _hists = []
    
def _layout_at_exit():
    from . import _master_show
    if _master_show:
        #print("here")
        from .layout import _run
        _run()
        _trigger()

register(_layout_at_exit)


from . import _config_dir, _trigger_file
import os

"""
Creates a file that triggers the vscode extension

"""
def _trigger():
    global _config_dir, _trigger_file
    #print("triggering")
    if _config_dir in os.listdir():
        with open(os.path.join(_config_dir, _trigger_file), 'w'): pass


"""
prepends inputted path with the config directory if it exists

"""
def get_path(path:str):
    global _config_dir
    if _config_dir in os.listdir(): 
        if os.path.isdir(_config_dir): return os.path.join(_config_dir, path)
    else: return path





from setuptools import setup
setup(
    name='U_Cincy_quantum_tools',
    version='0.0.2',
    author='Marek Brodke, with support from the University of Cincinnati',
    description='Provides functionaliy for UC_Quantum_Lab development tools',
    long_description="Provides functionaliy for UC_Quantum_Lab development tools",
    keywords='development',
    python_requires='>=3.7',
    license="MIT",
    author_email="brodkemd@mail.uc.edu",
    url="https://github.com/brodkemd/UCQ_tools",
    install_requires=[
        'qiskit>=0.36',
        'matplotlib>=2.2.0',
        'qiskit-aer>=0.10.4',
        'qiskit-ibmq-provider>=0.19.1',
        'qiskit-ignis>=0.7.0',
        'qiskit-terra>=0.20.1',
        "pylatexenc>=2.0"
    ]
)